import requests
import logging

from django.http import HttpResponseForbidden, JsonResponse
from django.http.response import HttpResponse
from django.core.paginator import Paginator
from django.shortcuts import redirect
from django.utils.text import slugify
from django.core.cache import cache
from django.utils import timezone
from django.db import connection
from django.conf import settings
from django.db.models import Q
from django.views import View

from dnoticias_services.mail import send_email

logger = logging.getLogger(__name__)


class Select2View(View):
    RESULTS = "results"
    TEXT = "text"
    CHILDREN = "children"
    IMAGE = "img"
    ID = "id"

    TERM_QS = "term"
    Q_QS = "q"
    _TYPE_QS = "_type"
    PAGE_QS = "page"

    PAGINATION_SIZE = 20
    PAGINATION = "pagination"
    MORE = "more"

    SEARCH_PARAMS = [] # Need to override
    SEARCH_TYPE = "" # Need to override
    ORDER_BY_PARAMS = [] # Need to override
    MODEL = None # Need to override
    MODEL_DEFAULT_MANAGER = "_default_manager"
    MANAGER = MODEL_DEFAULT_MANAGER
    MODEL_VERBOSE_NAMES = {} # Need to override

    def get(self, request, *args, **kwargs):
        self.term = request.GET.get(self.TERM_QS, "")
        try:
            self.page = int(request.GET.get(self.PAGE_QS, 1))
        except Exception as e:
            self.page = 1

        default_manager = getattr(self.MODEL, self.MODEL_DEFAULT_MANAGER, None)
        model_manager = getattr(self.MODEL, self.MANAGER, default_manager)
        if not model_manager:
            model_manager = default_manager

        elements = model_manager.all()
        elements = self.get_initial_filter(elements, request, *args, **kwargs)
        if self.term:
            elements = self.filter_by_term(elements)

        group_by_context = self.group_elements_by_context(elements)

        has_more = []
        results = []
        for key in group_by_context:
            children = []
            paginator = Paginator(group_by_context[key], self.PAGINATION_SIZE)
            page = paginator.page(self.page)
            has_more.append(page.has_next())
            for element in page.object_list:
                element_context = {
                    self.ID : element.pk
                }
                str_element = self.stringify_element(element)
                if str_element:
                    element_context[self.TEXT] = str_element

                element_context.update(self.get_extra_context(element))
                children.append(element_context)
            results.append({
                self.TEXT : key,
                self.CHILDREN : children
            })

        response_context = {
            self.RESULTS : results,
            self.PAGINATION: {
                self.MORE: any(has_more)
            }
        }
        return JsonResponse(data=response_context)

    def get_extra_context(self, element):
        return {}

    def stringify_element(self, element):
        if element:
            return str(element)
        else:
            return None

    def group_elements_by_context(self, elements):
        group_by_context = {}
        for element in elements:
            model_name = element.__class__.__name__
            verbose_name = self.MODEL_VERBOSE_NAMES.get(model_name, model_name)
            if verbose_name in group_by_context:
                group_by_context[verbose_name].append(element)
            else:
                group_by_context[verbose_name] = [element]
        return group_by_context

    def filter_by_term(self, elements):
        qs = None
        filter_context = dict()
        for param in self.SEARCH_PARAMS:
            filter_context = {
                param + "__" + self.SEARCH_TYPE : self.term
            }
            if not qs:
                qs = Q(**filter_context)
                continue

            qs |= Q(**filter_context)
        return elements.filter(qs).order_by(*self.ORDER_BY_PARAMS)

    def get_initial_filter(self, queryset, request, *args, **kwargs):
        return queryset

class SlugifyView(View):
    MODEL = None

    def get(self, request, *args, **kwargs):
        title = request.GET.get("title", "")
        pk = request.GET.get("pk", None)
        pk = pk if pk else None
        return JsonResponse(data={
            "slug" : self.get_unique_slug(title=title, pk=pk)
        })

    def get_unique_slug(title=None, pk=None):
        logger.debug("Transforming {} into a slug...".format(title))

        assert (title is not None and pk is not None)

        pk = pk if not bool(instance) else instance.pk
        title = title if not bool(instance) else instance.title
        slug = slugify(title)

        queryset = self.MODEL._default_manager.filter(~Q(pk=pk), slug=slug)
        if queryset:
            if pk:
                identifier = pk
            else:
                identifier = queryset.last().pk + 1 if queryset.exists() else 1
            slug = "{}-{}".format(slug, identifier)
        logger.debug("{} into a slug -> {}".format(title, slug))
        return slug

class AjaxSlugifyView(SlugifyView):
    def has_authorization(self, request, *args, **kwargs):
        return True

    def dispatch(self, request, *args, **kwargs):
        if request.is_ajax():
            if self.has_authorization(request, *args, **kwargs):
                return super().dispatch(request, *args, **kwargs)
            else:
                return HttpResponseForbidden("Not enough permissions to request slug via ajax")
        else:
            return HttpResponseBadRequest

class AjaxView(View):
    def dispatch(self, request, *args, **kwargs):
        if self.request.is_ajax():
            return super().dispatch(request, *args, **kwargs)
        else:
            return HttpResponseBadRequest()

class GenericDeleteView(View):
    model = None

    def has_authorization(self, request, instance, *args, **kwargs):
        return True

    def post(self, request, *args, **kwargs):
        pk = kwargs.pop("pk")
        redirect_url = request.POST.get("redirect")

        instance = self.model._default_manager.get(pk=pk)
        if not self.has_authorization(request, instance, *args, **kwargs):
            return HttpResponseForbidden("No authorization to delete")

        self.process_delete(instance)

        response_context = {
            "result" : "OK"
        }
        extra_response_context = self.get_extra_response_context()
        response_context.update(extra_response_context)

        if redirect_url:
            self.add_message(request)
            return redirect(redirect_url)
        else:
            return JsonResponse(
                response_context,
                safe=False
            )

    def process_delete(self, instance):
        instance.delete()

    def add_message(self, request):
        pass

    def get_extra_response_context(self):
        return {}


class LivenessCheckView(View):
    """
    Checks the liveness doing two simple tests:
    1. Ensure the connection with database
    2. Making a request to homepage and check the response code
    """
    def _is_error(self, status_code):
        return bool(400 <= status_code < 600)

    def get(self, *args, **kwargs):
        status_code = 200
        context = {
            'error_at': timezone.now().strftime("%Y-%m-%d %H:%M:%S"),
            'base_url': settings.BASE_URL
        }

        # ensure_connection returns None if everything is OK, else throws an exception
        try:
            connection.ensure_connection()
        except Exception as exception:
            status_code = 500
            context['error'] = '{0}'.format(exception)
            logger.exception(
                "Cannot connect to database in health check {}!".format(timezone.now())
            )

        # Test the homepage response if LIVENESS_CHECK_HOMEPAGE is True and the previous test
        # was OK (status 200)
        if settings.LIVENESS_CHECK_HOMEPAGE and not self._is_error(status_code):
            try:
                response = requests.get(settings.BASE_URL)
            except Exception as exception:
                context['error'] = 'Homepage check returned status code {}'.format(status_code)

        # If LIVENESS_SEND_EMAIL_ON_FAILURE is True and the status code is not 200
        # then we send an email to the IT team communicating the failure
        print(self._is_error(status_code) and settings.LIVENESS_SEND_EMAIL_ON_FAILURE)
        if self._is_error(status_code) and settings.LIVENESS_SEND_EMAIL_ON_FAILURE:

            cache.set(settings.LIVENESS_CACHE_NAME, settings.LIVENESS_EMAIL_COOLDOWN)
            context["error_code"] = status_code

            send_email(
                settings.DEFAULT_IT_EMAIL,
                settings.EMAIL_TEMPLATE_LIVENESS_FAIL_UUID,
                settings.EMAIL_BRAND_GROUP_UUID,
                "FATAL: Liveness check failure on {}!".format(settings.BASE_URL),
                context=context,
            )

            logger.info("Email notification sended to {}".format(settings.DEFAULT_IT_EMAIL))

        return HttpResponse('', status=status_code)



# -*- coding: utf-8 -*-
#
# http://djangosnippets.org/snippets/1550/
#

import pdb as pdb_module

from django.template import Library, Node

register = Library()


class PdbNode(Node):

    def render(self, context):
        pdb_module.set_trace()
        return ''


@register.tag
def pdb(parser, token):
    return PdbNode()



import importlib
import re
import logging
import traceback
from itertools import chain
from urllib.parse import urlparse, parse_qs, urlencode, urlsplit, urlunsplit

from django import forms
from django.contrib import messages
from django.shortcuts import reverse
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.models import Manager, ManyToManyField
from django.template.loader import get_template, TemplateDoesNotExist, render_to_string

logger = logging.getLogger(__name__)


def normalize_domain(domain):
    return re.sub('[^\w\-_ ]', '', domain)


def get_model_attr(instance, field, default=[]):
    try:
        value = getattr(instance, field, None)
    except Exception as e:
        logger.exception(e)
        value = None
    is_manager = isinstance(value, Manager)
    if is_manager:
        value = value.all()
    return value


def get_attr_by_string(class_path):
    path_fragments = class_path.split(".")
    module_path = ".".join(path_fragments[:-1])
    class_name = path_fragments[-1]
    module = importlib.import_module(module_path)
    _class = getattr(module, class_name, None)
    if _class:
        return _class
    else:
        raise Exception(
            "Couldn't find the class with the path {}".format(class_path))


def exists_template(template_name):
    try:
        get_template(template_name)
        return True
    except TemplateDoesNotExist as e:
        return False


def render_to_string_or_empty(template_name, context, request=None, using=None):
    logger.debug("TEMPLATE NAME -> {}".format(template_name))
    try:
        content = render_to_string(
            template_name, context=context, request=request, using=using
        )
    except Exception as e:
        logger.exception("Error in rendering the template -> {}".format(template_name))
        is_preview = context.get("is_preview")
        if is_preview:
            content = traceback.format_exc()
        else:
            content = ""
    return content


def instance_to_dot_path(instance):
    if isinstance(instance, type):
        return ".".join([instance.__module__, instance.__name__])
    else:
        return ".".join([instance.__class__.__module__, instance.__class__.__name__])


def errors_serialize(form):
    errors_context = dict()
    for key in form.errors:
        if key in form:
            errors_context[key] = {
                "id": form[key].auto_id,
                "name": form[key].html_name,
                "errors": form.errors[key]
            }
    return errors_context


def to_snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def get_facebook_graph_url():
    facebook_graph_base_url = settings.FACEBOOK_GRAPH_URL
    if facebook_graph_base_url.endswith('/'):
        facebook_graph_base_url += "/"
    return facebook_graph_base_url


def is_absolute(url):
    return bool(urlparse(url).netloc)


def get_field_form_from_model(instance, field_name):
    instance_class = instance.__class__

    class CustomModelForm(forms.ModelForm):
        class Meta:
            model = instance_class
            fields = (field_name,)
    form = CustomModelForm(instance=instance)
    return form.fields[field_name]


def to_valid_variable_name(value):
    # Remove invalid characters
    value = re.sub('[^0-9a-zA-Z_]', '_', value)

    # Remove leading characters until we find a letter or underscore
    value = re.sub('^[^a-zA-Z_]+', '', value)

    return value


def replace_keys_with_valid_variable_name(context):
    new_context = dict()
    keys = context.keys()
    for key in keys:
        new_context[to_valid_variable_name(key)] = context.get(key)
    return new_context


def get_content_type_preview_url(printable_object):
    if printable_object:
        return reverse("preview-content-type", kwargs={
            "content_type_pk" : ContentType.objects.get_for_model(printable_object).pk,
            "pk" : printable_object.pk
        })
    else:
        return None


def send_message(request, msg_level, msg, **kwargs):
    fail_silently = kwargs.pop("fail_silently", False)

    if fail_silently:
        try:
            messages.add_message(request, msg_level, msg, **kwargs)
        except Exception as e:
            logger.exception("Failed to add message to the request -> {}, probably request is None".format(request))
    else:
        messages.add_message(request, msg_level, msg, **kwargs)


def set_query_parameters(url, **extra_query_params):
    scheme, netloc, path, query_string, fragment = urlsplit(url)
    query_params = parse_qs(query_string)

    for param in extra_query_params:
        query_params[param] = extra_query_params[param]

    new_query_string = urlencode(query_params, doseq=True)

    return urlunsplit((scheme, netloc, path, new_query_string, fragment))


def has_cached_prefetch_field(model_instance, attr_name):
    attr_value = getattr(model_instance, attr_name, None)
    if attr_value and isinstance(attr_value, list):
        return True

    return bool(
        getattr(model_instance, "_prefetched_objects_cache", {}).get(attr_name)
    )


def has_cached_related_field(model_instance, model_field=None, attr_name=None):
    if model_field is None and attr_name is None:
        return False

    field_name = model_field.field.name if model_field else attr_name

    return bool(
        getattr(model_instance._state, "fields_cache", {}).get(field_name)
    )


def get_select_related_choices_by_model(model_class):
    opts = model_class._meta

    direct_choices = (f.name for f in opts.fields if f.is_relation)

    reverse_choices = (
        f.field.related_query_name()
        for f in opts.related_objects if f.field.unique
    )

    return list(
        chain(direct_choices, reverse_choices)
    )


def get_prefetch_related_choices_by_model(model_class):
    opts = model_class._meta

    reverse_choices = (
        f.field.related_query_name()
        for f in opts.related_objects if not f.field.unique
    )

    direct_choices = (f.name for f in opts.get_fields() if isinstance(f, ManyToManyField))

    return list(
        chain(
            reverse_choices,
            direct_choices
        )
    )



from django.template import Library
from django.utils import timezone
from datetime import datetime

import logging

logger = logging.getLogger(__name__)

register = Library()

@register.simple_tag
def current_timestamp():
    now = timezone.now()
    return int(datetime.timestamp(now))



from django.shortcuts import render


def set_error_context(description, error_code):
    return {
        'error_code': error_code,
        'description': description,
    }


def handler_500(request):
    context = set_error_context("Upsss, houve um erro no pedido. Tente mais tarde.", 500)
    return render(request, "error.html", context=context, status=500)


def handler_404(request, exception):
    context = set_error_context("A página solicitada não foi encontrada.", 404)
    return render(request, "error.html", context=context, status=404)


def handler_403(request, exception):
    context = set_error_context("Eiii, esta página não é para ti!", 403)
    return render(request, "error.html", context=context, status=403)


def handler_400(request, exception):
    context = set_error_context("O servidor não conseguiu interpretar o seu pedido.", 400)
    return render(request, "error.html", context=context, status=400)



from django.template import Library, Node
import math

register = Library()

@register.simple_tag
def rest_of_division(a, b):
    return a % b

@register.simple_tag
def division(a, b):
    return a / b

@register.simple_tag
def ceil_division(a, b):
    return math.ceil(a / b)

@register.simple_tag
def sum(a, b):
    return int(a) + int(b)

@register.simple_tag
def trange(*args):
    return range(*args)

@register.simple_tag
def subtract(a , b):
    return a - b
